package com.maddyhome.idea.color;

/*
 * Color Browser - Color browser plugin for IDEA
 * Copyright (C) 2006 Rick Maddy. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import com.intellij.codeInsight.CodeInsightColors;
import com.intellij.codeInsight.hint.HintUtil;
import com.intellij.openapi.components.ApplicationComponent;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.EditorFactory;
import com.intellij.openapi.editor.LogicalPosition;
import com.intellij.openapi.editor.event.EditorFactoryAdapter;
import com.intellij.openapi.editor.event.EditorFactoryEvent;
import com.intellij.openapi.editor.event.EditorMouseEvent;
import com.intellij.openapi.editor.event.EditorMouseListener;
import com.intellij.openapi.editor.event.EditorMouseMotionListener;
import com.intellij.openapi.editor.event.VisibleAreaEvent;
import com.intellij.openapi.editor.event.VisibleAreaListener;
import com.intellij.openapi.editor.markup.HighlighterLayer;
import com.intellij.openapi.editor.markup.HighlighterTargetArea;
import com.intellij.openapi.editor.markup.MarkupModel;
import com.intellij.openapi.editor.markup.RangeHighlighter;
import com.intellij.openapi.editor.markup.TextAttributes;
import com.intellij.openapi.fileEditor.FileEditor;
import com.intellij.openapi.fileEditor.FileEditorManager;
import com.intellij.openapi.fileEditor.FileEditorManagerAdapter;
import com.intellij.openapi.fileEditor.FileEditorManagerEvent;
import com.intellij.openapi.fileEditor.TextEditor;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.project.ProjectManager;
import com.intellij.openapi.project.ProjectManagerAdapter;
import com.intellij.openapi.util.Key;
import com.maddyhome.idea.color.search.SearchEngine;
import com.maddyhome.idea.color.util.ColorInfo;
import com.maddyhome.idea.color.util.MatchRange;
import org.jetbrains.annotations.NonNls;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.util.Map;
import java.util.WeakHashMap;
import javax.swing.JLabel;
import javax.swing.Popup;
import javax.swing.PopupFactory;
import javax.swing.SwingUtilities;

/**
 * This plugin handles the Ctrl-Mouse Move functionality in an editor. If the cursor is over a valid color
 * representation a popup is displayed showing the color.
 */
public class ColorInfoPluginImpl implements ApplicationComponent
{
    @NonNls
    public String getComponentName()
    {
        return "ColorInfo";
    }

    public void initComponent()
    {
        EditorFactory editorFactory = EditorFactory.getInstance();
        editorFactory.addEditorFactoryListener(editorListener);

        // If the user switches the current editor then clear any highlights
        ProjectManager.getInstance().addProjectManagerListener(new ProjectManagerAdapter()
        {
            public void projectOpened(Project project)
            {
                FileEditorManager.getInstance(project).addFileEditorManagerListener(new FileEditorManagerAdapter()
                {
                    public void selectionChanged(FileEditorManagerEvent e)
                    {
                        FileEditor feditor = e.getOldEditor();
                        if (feditor instanceof TextEditor)
                        {
                            Editor editor = ((TextEditor)feditor).getEditor();
                            clearHighlighters(editor);
                        }
                    }
                });
            }
        });

        defaultHint = HintUtil.createInformationLabel("x");
    }

    public void disposeComponent()
    {
        EditorFactory editorFactory = EditorFactory.getInstance();
        editorFactory.removeEditorFactoryListener(editorListener);
    }

    private MatchRange getHyperLinkRange(Editor editor)
    {
        LogicalPosition logicalPosition = editor.xyToLogicalPosition(lastMouseLocation);
        Document document = editor.getDocument();
        int offset = editor.logicalPositionToOffset(logicalPosition);
        int length = document.getTextLength();
        if (length > 0 && offset < length)
        {
            int start = document.getLineStartOffset(logicalPosition.line);
            int end = document.getLineEndOffset(logicalPosition.line);
            int pos = offset - start;
            CharSequence text = document.getCharsSequence().subSequence(start, end + 1);

            return SearchEngine.findColorMatch(text.toString(), pos, start);
        }

        return null;
    }

    private boolean setHighlighter(Editor editor, MatchRange range)
    {
        MatchRange current = editor.getUserData(LINKRANGE);
        if (!range.equals(current))
        {
            Color color = SearchEngine.getColor(range.getMatch());
            if (color == null)
            {
                return false;
            }

            TextAttributes attrs = editor.getColorsScheme().getAttributes(CodeInsightColors.HYPERLINK_ATTRIBUTES);
            MarkupModel markupModel = editor.getMarkupModel();
            RangeHighlighter highlighter = editor.getUserData(HIGHLIGHTER);
            if (highlighter != null)
            {
                markupModel.removeHighlighter(highlighter);
            }
            highlighter = markupModel.addRangeHighlighter(
                range.getStartOffset(), range.getEndOffset(),
                HighlighterLayer.ADDITIONAL_SYNTAX, attrs,
                HighlighterTargetArea.EXACT_RANGE
            );
            editor.putUserData(HIGHLIGHTER, highlighter);
            editor.putUserData(LINKRANGE, range);

            clearPopup(editor);

            Popup popup = createPopup(editor, color);
            popup.show();
            editor.putUserData(POPUP, popup);
        }

        return true;
    }

    private Popup createPopup(final Editor editor, Color color)
    {
        JLabel label = new JLabel();
        label.setBorder(defaultHint.getBorder());
        label.setOpaque(defaultHint.isOpaque());
        label.setBackground(defaultHint.getBackground());
        label.setForeground(defaultHint.getForeground());
        label.setText(createColorText(defaultHint.getFont(), color));

        label.addMouseMotionListener(new MouseMotionListener()
        {
            public void mouseDragged(MouseEvent e)
            {
            }

            public void mouseMoved(MouseEvent e)
            {
                clearPopup(editor);
            }
        });

        LogicalPosition lp = editor.xyToLogicalPosition(lastMouseLocation);
        Point pos = editor.logicalPositionToXY(lp);
        pos.x = lastMouseLocation.x;
        pos.y -= label.getPreferredSize().height;
        SwingUtilities.convertPointToScreen(pos, editor.getContentComponent());
        Rectangle rect = editor.getComponent().getRootPane().getBounds();
        Point corner = new Point(rect.x + rect.width, 0);
        SwingUtilities.convertPointToScreen(corner, editor.getComponent().getRootPane());
        if (pos.x + label.getPreferredSize().width > corner.x)
        {
            pos.x = Math.max(0, corner.x - label.getPreferredSize().width);
        }
        return PopupFactory.getSharedInstance().getPopup(editor.getContentComponent(), label, pos.x, pos.y);
    }

    private String createColorText(Font font, Color color)
    {
        ColorInfo info = new ColorInfo(color);
        StringBuffer res = new StringBuffer();

        res.append("<html><body style=\"font: ");
        res.append(font.getSize());
        res.append("pt ");
        res.append(font.getFamily());
        res.append("\">");
        res.append("<table border=\"1\" cellspacing=\"0\" cellpadding=\"0\">");
        res.append(  "<tr>");
        res.append(    "<td style=\"border: 1 solid black; background-color: ");
        res.append(info.getHex());
        res.append("; width: 50;\">");
        res.append(      "&nbsp;");
        res.append(    "</td>");
        res.append(    "<td>");
        res.append(      "<table border=\"0\" cellspacing=\"3\" cellpadding=\"0\">");
        res.append(        "<tr>");
        res.append(          "<td>");
        res.append(            "<l>Hex</l>:");
        res.append(          "</td>");
        res.append(          "<td>");
        res.append(            info.getHex());
        res.append(          "</td>");
        res.append(        "</tr>");
        res.append(        "<tr>");
        res.append(          "<td>");
        res.append(            "<l>Decimal</l>:");
        res.append(          "</td>");
        res.append(          "<td>");
        res.append(            info.getDecimalRGB());
        res.append(          "</td>");
        res.append(        "</tr>");
        res.append(      "</table>");
        res.append(    "</td>");
        res.append(  "</tr>");
        res.append("</table>");
        res.append("</body></html>");

        return res.toString();
    }

    private void clearHighlighters(Editor editor)
    {
        RangeHighlighter highlighter = editor.getUserData(HIGHLIGHTER);
        if (highlighter != null)
        {
            MarkupModel markupModel = editor.getMarkupModel();
            markupModel.removeHighlighter(highlighter);
            editor.putUserData(HIGHLIGHTER, null);
            editor.putUserData(LINKRANGE, null);
            editor.getContentComponent().setToolTipText(null);
        }

        clearPopup(editor);
    }

    private void clearPopup(Editor editor)
    {
        Popup popup = editor.getUserData(POPUP);
        if (popup != null)
        {
            popup.hide();
            editor.putUserData(POPUP, null);
        }
    }

    private class EditorListener extends EditorFactoryAdapter
    {
        private final Map<Editor, ControlHandler> keyHandlers = new WeakHashMap<Editor, ControlHandler>();

        public void editorCreated(EditorFactoryEvent event)
        {
            Editor editor = event.getEditor();
            ControlHandler controlHandler = new ControlHandler(editor);
            keyHandlers.put(editor, controlHandler);
            editor.addEditorMouseMotionListener(controlHandler);
            editor.addEditorMouseListener(controlHandler);
            editor.getContentComponent().addKeyListener(controlHandler);
            editor.getScrollingModel().addVisibleAreaListener(controlHandler);
        }

        public void editorReleased(EditorFactoryEvent event)
        {
            Editor editor = event.getEditor();
            ControlHandler controlHandler = keyHandlers.remove(editor);
            editor.removeEditorMouseMotionListener(controlHandler);
            editor.removeEditorMouseListener(controlHandler);
            editor.getContentComponent().removeKeyListener(controlHandler);
            editor.getScrollingModel().removeVisibleAreaListener(controlHandler);
        }
    }

    private class ControlHandler implements KeyListener, EditorMouseMotionListener, EditorMouseListener, VisibleAreaListener
    {
        public ControlHandler(Editor editor)
        {
            this.editor = editor;
        }

        public void keyTyped(KeyEvent e)
        {
            clearPopup(editor);
        }

        public void keyPressed(KeyEvent e)
        {
            clearPopup(editor);
        }

        public void keyReleased(KeyEvent e)
        {
            if ((e.getModifiersEx() & KeyEvent.CTRL_DOWN_MASK) != KeyEvent.CTRL_DOWN_MASK)
            {
                clearHighlighters(editor);

                editor.getContentComponent().setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
            }
        }

        public void mouseMoved(EditorMouseEvent e)
        {
            MouseEvent mouseEvent = e.getMouseEvent();
            lastMouseLocation.setLocation(mouseEvent.getPoint());
            if ((e.getMouseEvent().getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) == MouseEvent.CTRL_DOWN_MASK)
            {
                Editor editor = e.getEditor();
                MatchRange range = getHyperLinkRange(editor);
                if (range != null)
                {
                    if (setHighlighter(editor, range))
                    {
                        editor.getContentComponent().setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
                    }
                }
            }
        }

        public void mouseDragged(EditorMouseEvent e)
        {
            // no-op
        }

        public void mousePressed(EditorMouseEvent e)
        {
            // no-op
        }

        public void mouseClicked(EditorMouseEvent e)
        {
            if (!e.isConsumed() &&
                (e.getMouseEvent().getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) == MouseEvent.CTRL_DOWN_MASK &&
                e.getMouseEvent().getButton() == MouseEvent.BUTTON1)
            {
                Editor editor = e.getEditor();
                MatchRange range = getHyperLinkRange(editor);
                if (range != null)
                {
                    Color color = SearchEngine.getColor(range.getMatch());
                    if (color != null)
                    {
                        Project project = editor.getProject();
                        if (project != null)
                        {
                            ColorBrowserPlugin plugin = project.getComponent(ColorBrowserPlugin.class);
                            plugin.showAtCursor();
                        }
                    }
                }
            }

            clearHighlighters(e.getEditor());
        }

        public void mouseReleased(EditorMouseEvent e)
        {
            // no-op
        }

        public void mouseEntered(EditorMouseEvent e)
        {
            // no-op
        }

        public void mouseExited(EditorMouseEvent e)
        {
            // no-op
        }

        public void visibleAreaChanged(VisibleAreaEvent visibleAreaEvent)
        {
            clearHighlighters(editor);
        }

        private final Editor editor;
    }

    private static final Key<RangeHighlighter> HIGHLIGHTER = Key.create("ColorLinkHighlighter");
    private static final Key<MatchRange> LINKRANGE = Key.create("ColorLinkRange");
    private static final Key<Popup> POPUP = Key.create("ColorLinkPopup");

    private final EditorListener editorListener = new EditorListener();
    private final Point lastMouseLocation = new Point(0, 0);
    private JLabel defaultHint;
}