package com.maddyhome.idea.color.search;

/*
 * Color Browser - Color browser plugin for IDEA
 * Copyright (C) 2006 Rick Maddy. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import com.intellij.openapi.diagnostic.Logger;
import com.maddyhome.idea.color.config.Configuration;
import com.maddyhome.idea.color.util.ColorUtil;
import com.maddyhome.idea.color.util.EditorSelection;
import com.maddyhome.idea.color.util.MatchRange;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.awt.Color;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class SearchEngine
{
    @Nullable
    public static MatchRange findColorMatch(@NotNull String line, int pos, int offset)
    {
        if (pos > 0 && line.charAt(pos) == ';')
        {
            pos--;
        }

        for (Pattern pattern : COLOR_PATTERNS)
        {
            Matcher matcher = pattern.matcher(line);
            while (matcher.find())
            {
                logger.debug("matcher=" + matcher);
                if (matcher.start() <= pos && pos < matcher.end())
                {
                    return new MatchRange(matcher.start() + offset, matcher.end() + offset, matcher.group());
                }
            }
        }

        return null;
    }

    @NotNull
    public static SearchResults findColor(@NotNull EditorSelection selection)
    {
        String line = selection.getLine();
        int lstart = selection.getStart();
        int lend = selection.getEnd();

        if (lstart != lend)
        {
            String highlight = line.substring(lstart, lend);
            Color color = getColor(highlight);
            if (color != null)
            {
                SearchResults res = new SearchResults(selection);

                res.addColor(color);

                return res;
            }
        }

        SearchResults res = new SearchResults(selection);

        MatchRange range = findColorMatch(line, lstart, 0);
        if (range != null)
        {
            Color ch = getColor(range.getMatch());
            if (ch != null)
            {
                res.addColor(ch);
            }
        }

        return res;
    }

    @Nullable
    public static Color getColor(@NotNull String text)
    {
        logger.debug("text='" + text + "'");
        try
        {
            if (text.startsWith("#"))
            {
                if (text.length() == 4)
                {
                    return ColorUtil.getShortRGB(text.substring(1));
                }
                else
                {
                    return ColorUtil.getRGB(text.substring(1));
                }
            }
            else if (text.startsWith("rgb"))
            {
                boolean isPercent = false;
                float a = 1.0f;
                int r, g, b;
                int ps = text.indexOf('(');
                int pe = text.indexOf(')');
                if (ps == -1 || pe == -1)
                {
                    return null;
                }

                StringTokenizer tokenizer = new StringTokenizer(text.substring(ps + 1, pe), ",");
                String part = tokenizer.nextToken().trim();
                if (part.endsWith("%"))
                {
                    isPercent = true;
                    r = Integer.parseInt(part.substring(0, part.length() - 1));
                }
                else
                {
                    r = Integer.parseInt(part);
                }

                part = tokenizer.nextToken().trim();
                if (part.endsWith("%"))
                {
                    isPercent = true;
                    g = Integer.parseInt(part.substring(0, part.length() - 1));
                }
                else
                {
                    g = Integer.parseInt(part);
                }

                part = tokenizer.nextToken().trim();
                if (part.endsWith("%"))
                {
                    isPercent = true;
                    b = Integer.parseInt(part.substring(0, part.length() - 1));
                }
                else
                {
                    b = Integer.parseInt(part);
                }

                if (tokenizer.hasMoreTokens())
                {
                    part = tokenizer.nextToken().trim();
                    a = Float.parseFloat(part);
                }

                logger.debug("r=" + r);
                logger.debug("g=" + g);
                logger.debug("b=" + b);
                logger.debug("a=" + a);
                if (isPercent)
                {
                    return ColorUtil.getPercentRGBa(r, g, b, a);
                }
                else
                {
                    return ColorUtil.getDecimalRBGa(r, g, b, a);
                }
            }
            else if (text.startsWith("hsl"))
            {
                float a = 1.0f;
                int h, s, l;
                int ps = text.indexOf('(');
                int pe = text.indexOf(')');
                if (ps == -1 || pe == -1)
                {
                    return null;
                }

                StringTokenizer tokenizer = new StringTokenizer(text.substring(ps + 1, pe), ",");
                String part = tokenizer.nextToken().trim();
                h = Integer.parseInt(part);

                part = tokenizer.nextToken().trim();
                if (part.endsWith("%"))
                {
                    s = Integer.parseInt(part.substring(0, part.length() - 1));
                }
                else
                {
                    s = Integer.parseInt(part);
                }

                part = tokenizer.nextToken().trim();
                if (part.endsWith("%"))
                {
                    l = Integer.parseInt(part.substring(0, part.length() - 1));
                }
                else
                {
                    l = Integer.parseInt(part);
                }

                if (tokenizer.hasMoreTokens())
                {
                    part = tokenizer.nextToken().trim();
                    a = Float.parseFloat(part);
                }

                logger.debug("h=" + h);
                logger.debug("s=" + s);
                logger.debug("l=" + l);
                logger.debug("a=" + a);

                return ColorUtil.getHSLa(h, s, l, a);
            }
            else if (text.startsWith("Color."))
            {
                String name = text.substring("Color.".length());
                Color color = Configuration.getInstance().findJavaColor(name);
                if (color != null)
                {
                    return color;
                }
            }
            else if (text.startsWith("ColorUIResource."))
            {
                String name = text.substring("ColorUIResource.".length());
                Color color = Configuration.getInstance().findJavaColor(name);
                if (color != null)
                {
                    return color;
                }
            }
            else if (text.startsWith("Color") || text.startsWith("ColorUIResource"))
            {
                boolean isFloat = false;
                float fr = 0.0f, fg = 0.0f, fb = 0.0f;
                float fa = 1.0f;
                int ir = 0, ig = 0, ib = 0;
                int ia = 255;
                boolean alpha = false;
                int ps = text.indexOf('(');
                int pe = text.indexOf(')');
                if (ps == -1 || pe == -1)
                {
                    return null;
                }

                StringTokenizer tokenizer = new StringTokenizer(text.substring(ps + 1, pe), ",");
                int params = tokenizer.countTokens();
                String part = tokenizer.nextToken().trim();
                if (part.endsWith("f"))
                {
                    isFloat = true;
                    fr = Float.parseFloat(part.substring(0, part.length() - 1));
                }
                else
                {
                    ir = parseInt(part);
                }

                if (params >= 2)
                {
                    part = tokenizer.nextToken().trim();
                    if (part.equals("true"))
                    {
                        alpha = true;
                    }
                    else if (part.equals("false"))
                    {
                        alpha = false;
                    }
                    else if (part.endsWith("f"))
                    {
                        isFloat = true;
                        fg = Float.parseFloat(part.substring(0, part.length() - 1));
                    }
                    else
                    {
                        ig = parseInt(part);
                    }

                    if (params >= 3)
                    {
                        part = tokenizer.nextToken().trim();
                        if (part.endsWith("f"))
                        {
                            isFloat = true;
                            fb = Float.parseFloat(part.substring(0, part.length() - 1));
                        }
                        else
                        {
                            ib = parseInt(part);
                        }

                        if (params == 4)
                        {
                            part = tokenizer.nextToken().trim();
                            if (part.endsWith("f"))
                            {
                                isFloat = true;
                                fa = Float.parseFloat(part.substring(0, part.length() - 1));
                            }
                            else
                            {
                                ia = parseInt(part);
                            }
                        }
                    }
                }

                if (isFloat)
                {
                    return new Color(fr, fg, fb, fa);
                }
                else
                {
                    if (params == 1)
                    {
                        return new Color(ir);
                    }
                    else if (params == 2)
                    {
                        return new Color(ir, alpha);
                    }
                    else
                    {
                        return new Color(ir, ig, ib, ia);
                    }
                }
            }
            else
            {
                Color res = Configuration.getInstance().findSVGColor(text.toLowerCase());
                if (res != null)
                {
                    return res;
                }
            }
        }
        catch (NumberFormatException e)
        {
            logger.error(e);
            return null;
        }

        return null;
    }

    private static int parseInt(String part)
    {
        int res;

        if (part.toLowerCase().startsWith("0x"))
        {
            res = Integer.parseInt(part.substring(2), 16);
        }
        else if (part.startsWith("0") && part.length() > 1)
        {
            res = Integer.parseInt(part.substring(1), 8);
        }
        else
        {
            res = Integer.parseInt(part);
        }

        return res;
    }

    private SearchEngine() {}

    // The first set of patterns match characters
    // The second set of patterns match integer literals
    private static final Pattern[] COLOR_PATTERNS = new Pattern[]{
        Pattern.compile(
            "((#\\p{XDigit}{6}\\b)|(#\\p{XDigit}{3}\\b))"
        ),
        Pattern.compile(
            "\\b((rgb\\s*\\(\\s*\\p{Digit}{1,3}\\s*,\\s*\\p{Digit}{1,3}\\s*,\\s*\\p{Digit}{1,3}\\s*\\))|(rgb\\s*\\(\\s*\\p{Digit}{1,3}%\\s*,\\s*\\p{Digit}{1,3}%\\s*,\\s*\\p{Digit}{1,3}%\\s*\\)))"
        ),
        Pattern.compile(
            "\\b((rgba\\s*\\(\\s*\\p{Digit}{1,3}\\s*,\\s*\\p{Digit}{1,3}\\s*,\\s*\\p{Digit}{1,3}\\s*,\\s*[0-9.]{1,3}\\s*\\))|(rgba\\s*\\(\\s*\\p{Digit}{1,3}%\\s*,\\s*\\p{Digit}{1,3}%\\s*,\\s*\\p{Digit}{1,3}%\\s*,\\s*[0-9.]{1,3}\\s*\\)))"
        ),
        Pattern.compile(
            "\\b((hsl\\s*\\(\\s*\\p{Digit}{1,3}\\s*,\\s*\\p{Digit}{1,3}%\\s*,\\s*\\p{Digit}{1,3}%\\s*\\)))"
        ),
        Pattern.compile(
            "\\b((hsla\\s*\\(\\s*\\p{Digit}{1,3}\\s*,\\s*\\p{Digit}{1,3}%\\s*,\\s*\\p{Digit}{1,3}%\\s*,\\s*[0-9.]{1,3}\\s*\\)))"
        ),
        Pattern.compile(
            "\\b(Color\\s*\\(\\s*(0[xX])?[0-9a-fA-F]+\\s*\\))"
        ),
        Pattern.compile(
            "\\b(Color\\s*\\(\\s*(0[xX])?[0-9a-fA-F]+\\s*,\\s*(true|false)\\s*\\))"
        ),
        Pattern.compile(
            "\\b(Color\\s*\\(\\s*(0[xX])?[0-9a-fA-F]+\\s*,\\s*(0[xX])?[0-9a-fA-F]+\\s*,\\s*(0[xX])?[0-9a-fA-F]+\\s*\\))"
        ),
        Pattern.compile(
            "\\b(Color\\s*\\(\\s*(0[xX])?[0-9a-fA-F]+\\s*,\\s*(0[xX])?[0-9a-fA-F]+\\s*,\\s*(0[xX])?[0-9a-fA-F]+\\s*,\\s*(0[xX])?[0-9a-fA-F]+\\s*\\))"
        ),
        Pattern.compile(
            "\\b(Color\\s*\\(\\s*[0-9.]+f?\\s*,\\s*[0-9.]+f?\\s*,\\s*[0-9.]+f?\\s*\\))"
        ),
        Pattern.compile(
            "\\b(Color\\s*\\(\\s*[0-9.]+f?\\s*,\\s*[0-9.]+f?\\s*,\\s*[0-9.]+f?\\s*,\\s*[0-9.]+f?\\s*\\))"
        ),
        Pattern.compile(
            "\\b(Color\\.[a-zA-Z_]+)\\b"
        ),
        Pattern.compile(
            "\\b(ColorUIResource\\s*\\(\\s*[0-9.]+f?\\s*,\\s*[0-9.]+f?\\s*,\\s*[0-9.]+f?\\s*\\))"
        ),
        Pattern.compile(
            "\\b(ColorUIResource\\s*\\(\\s*(0[xX])?[0-9a-fA-F]+\\s*\\))"
        ),
        Pattern.compile(
            "\\b(ColorUIResource\\s*\\(\\s*(0[xX])?[0-9a-fA-F]+\\s*,\\s*(0[xX])?[0-9a-fA-F]+\\s*,\\s*(0[xX])?[0-9a-fA-F]+\\s*\\))"
        ),
        Pattern.compile(
            "\\b(ColorUIResource\\.[a-zA-Z_]+)\\b"
        ),
        Pattern.compile(
            "\\b([a-zA-Z]+)\\b"
        )
    };

    private static Logger logger = Logger.getInstance(SearchEngine.class.getName());
}