package com.maddyhome.idea.color.util;

/*
 * Color Browser - Color browser plugin for IDEA
 * Copyright (C) 2006 Rick Maddy. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import com.intellij.openapi.diagnostic.Logger;

import java.awt.Color;

public class ColorUtil
{
    public static Color getShortRGB(String rgb)
    {
        assert rgb.length() == 3;

        int r = Integer.parseInt(rgb.substring(0, 1), 16);
        int g = Integer.parseInt(rgb.substring(1, 2), 16);
        int b = Integer.parseInt(rgb.substring(2, 3), 16);

        return new Color(r, g, b);
    }

    public static Color getRGB(String rrggbb)
    {
        assert rrggbb.length() == 6;

        int r = Integer.parseInt(rrggbb.substring(0, 2), 16);
        int g = Integer.parseInt(rrggbb.substring(2, 4), 16);
        int b = Integer.parseInt(rrggbb.substring(4, 6), 16);

        return new Color(r, g, b);
    }

    public static Color getDecimalRBG(int r, int g, int b)
    {
        return getDecimalRBGa(r, g, b, 1.0f);
    }

    public static Color getDecimalRBGa(int r, int g, int b, float a)
    {
        r = normalizeDecimal(r);
        g = normalizeDecimal(g);
        b = normalizeDecimal(b);
        int x = toDecimal(normalizeFraction(a));

        return new Color(r, g, b, x);
    }

    public static Color getPercentRGB(int r, int g, int b)
    {
        return getPercentRGBa(r, g, b, 1.0f);
    }

    public static Color getPercentRGBa(int r, int g, int b, float a)
    {
        r = normalizePercent(r);
        g = normalizePercent(g);
        b = normalizePercent(b);
        a = normalizeFraction(a);

        return new Color((float)r / 100f, (float)g / 100f, (float)b / 100f, a);
    }

    public static Color getHSL(int h, int s, int l)
    {
        return getHSLa(h, s, l, 1.0f);
    }

    public static Color getHSLa(int h, int s, int l, float a)
    {
        h = normalizeDegrees(h);
        s = normalizePercent(s);
        l = normalizePercent(l);
        a = normalizeFraction(a);

        float[] rgb = convertHSLToRGB((float)h / 359f, (float)s / 100f, (float)l / 100f);

        return new Color(rgb[0], rgb[1], rgb[2], a);
    }

    public static int HSLtoRGB(float h, float s, float l)
    {
        float[] vals = convertHSLToRGB(h, s, l);

        int r = (int)(vals[0] * 255f + 0.5f);
        int g = (int)(vals[1] * 255f + 0.5f);
        int b = (int)(vals[2] * 255f + 0.5f);

        return (255 << 24) | (r << 16) | (g << 8) | b;
    }

    public static Color getHSLColor(float h, float s, float l)
    {
        return new Color(HSLtoRGB(h, s, l));
    }

    public static float[] RGBtoHSL(int r, int g, int b, float[] hsl)
    {
        if (hsl == null)
        {
            hsl = new float[3];
        }

        float[] vals = convertRGBToHSL(r, g, b);

        hsl[0] = vals[0];
        hsl[1] = vals[1];
        hsl[2] = vals[2];

        return hsl;
    }

    private static float[] convertHSLToRGB(float h, float s, float l)
    {
        float[] res = new float[3];

        /* From w3c.org
        HOW TO RETURN hsl.to.rgb(h, s, l):
            SELECT:
                l<=0.5: PUT l*(s+1) IN m2
                ELSE: PUT l+s-l*s IN m2
            PUT l*2-m2 IN m1
            PUT hue.to.rgb(m1, m2, h+1/3) IN r
            PUT hue.to.rgb(m1, m2, h    ) IN g
            PUT hue.to.rgb(m1, m2, h-1/3) IN l
            RETURN (r, g, l)

        HOW TO RETURN hue.to.rgb(m1, m2, h):
            IF h<0: PUT h+1 IN h
            IF h>1: PUT h-1 IN h
            IF h*6<1: RETURN m1+(m2-m1)*h*6
            IF h*2<1: RETURN m2
            IF h*3<2: RETURN m1+(m2-m1)*(2/3-h)*6
            RETURN m1
        */

        float m2 = (l <= 0.5f) ? l * (s + 1.0f) : l + s - l * s;
        float m1 = l * 2.0f - m2;

        res[0] = convertHueToRGB(m1, m2, h + 1.0f / 3.0f);
        res[1] = convertHueToRGB(m1, m2, h);
        res[2] = convertHueToRGB(m1, m2, h - 1.0f / 3.0f);

        return res;
    }

    private static float convertHueToRGB(float m1, float m2, float h)
    {
        if (h < 0.0f) h++;
        if (h > 1.0f) h--;
        if (h * 6.0f < 1.0f) return m1 + (m2 - m1) * h * 6.0f;
        if (h * 2.0f < 1.0f) return m2;
        if (h * 3.0f < 2.0f) return m1 + (m2 - m1) * (2.0f / 3.0f - h) * 6.0f;
        return m1;
    }

    private static float[] convertRGBToHSL(int r, int g, int b)
    {
        float[] res = new float[3];

        /*
        var_R = ( R / 255 )                     //Where RGB values = 0  255
        var_G = ( G / 255 )
        var_B = ( B / 255 )

        var_Min = min( var_R, var_G, var_B )    //Min. value of RGB
        var_Max = max( var_R, var_G, var_B )    //Max. value of RGB
        del_Max = var_Max - var_Min             //Delta RGB value

        L = ( var_Max + var_Min ) / 2

        if ( del_Max == 0 )                     //This is a gray, no chroma...
        {
           H = 0                                //HSL results = 0  1
           S = 0
        }
        else                                    //Chromatic data...
        {
           if ( L < 0.5 ) S = del_Max / ( var_Max + var_Min )
           else           S = del_Max / ( 2 - var_Max - var_Min )

           del_R = ( ( ( var_Max - var_R ) / 6 ) + ( del_Max / 2 ) ) / del_Max
           del_G = ( ( ( var_Max - var_G ) / 6 ) + ( del_Max / 2 ) ) / del_Max
           del_B = ( ( ( var_Max - var_B ) / 6 ) + ( del_Max / 2 ) ) / del_Max

           if      ( var_R == var_Max ) H = del_B - del_G
           else if ( var_G == var_Max ) H = ( 1 / 3 ) + del_R - del_B
           else if ( var_B == var_Max ) H = ( 2 / 3 ) + del_G - del_R

           if ( H < 0 ) ; H += 1
           if ( H > 1 ) ; H -= 1
        }
        */

        float vR = r / 255.0f;
        float vG = g / 255.0f;
        float vB = b / 255.0f;

        float min = Math.min(vR, Math.min(vG, vB));
        float max = Math.max(vR, Math.max(vG, vB));
        float delMax = max - min;

        float l = (max + min) / 2.0f;
        float h;
        float s;

        if (delMax == 0.0f)
        {
            h = 0.0f;
            s = 0.0f;
        }
        else
        {
            if (l < 0.5f)
            {
                s = delMax / (max + min);
            }
            else
            {
                s = delMax / ( 2.0f - max - min);
            }

            float dR = (((max - vR) / 6.0f) + (delMax / 2.0f)) / delMax;
            float dG = (((max - vG) / 6.0f) + (delMax / 2.0f)) / delMax;
            float dB = (((max - vB) / 6.0f) + (delMax / 2.0f)) / delMax;

            if (vR == max)
            {
                h = dB - dG;
            }
            else if (vG == max)
            {
                h = (1.0f/3.0f) + dR - dB;
            }
            else /* if (vB == max) */
            {
                h = (2.0f/3.0f) + dG - dR;
            }

            if (h < 0.0f) h++;
            if (h > 1.0f) h--;
        }

        res[0] = h;
        res[1] = s;
        res[2] = l;

        return res;
    }

    private static int normalizeDecimal(int x)
    {
        return Math.max(0, Math.min(x, 255));
    }

    private static int normalizePercent(int x)
    {
        return Math.max(0, Math.min(x, 100));
    }

    private static int normalizeDegrees(int x)
    {
        return Math.max(0, Math.min(x % 360, 359));
    }

    private static float normalizeFraction(float f)
    {
        return Math.max(0.0f, Math.max(f, 1.0f));
    }

    private static int toDecimal(float f)
    {
        return (int)(f * 255f + 0.5f);
    }

    public static void test()
    {
        int min = Integer.MAX_VALUE;
        int max = Integer.MIN_VALUE;

        logger.debug("start hsb");
        for (int h = 0; h < 360; h++)
        {
            for (int s = 0; s <= 100; s++)
            {
                for (int b = 0; b <= 100; b++)
                {
                    int col = Color.HSBtoRGB(h / 359f, s / 100f, b / 100f);
                    min = Math.min(min, col);
                    max = Math.max(max, col);
                }
            }
        }
        logger.debug("end hsb. min=" + Integer.toHexString(min) + ", max=" + Integer.toHexString(max));

        min = Integer.MAX_VALUE;
        max = Integer.MIN_VALUE;

        logger.debug("start hsl");
        for (int h = 0; h < 360; h++)
        {
            for (int s = 0; s <= 100; s++)
            {
                for (int b = 0; b <= 100; b++)
                {
                    int col = HSLtoRGB(h / 359f, s / 100f, b / 100f);
                    min = Math.min(min, col);
                    max = Math.max(max, col);
                }
            }
        }
        logger.debug("end hsl. min=" + Integer.toHexString(min) + ", max=" + Integer.toHexString(max));
    }

    private ColorUtil() {}

    private static Logger logger = Logger.getInstance(ColorUtil.class.getName());
}
